.wy-nav-content {
    max-width: 1200px !important;
}

.green {
	color: green;
}

.statusok {
	color: green;
}

.statusmok {
	/* color: orange; */
	/* color: #ffae42; */
	color: darkorange;
}

.statusunst {
	color: red;
}
